# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import DB, OptionalC, OptionalNC, NotifyTable
from tools import formats
from tools.DataModel import DataModel, Id, Text, Int, Date, RefProp
from tools.DataModel import InvalidAttributeError, MismatchROError, MissingRequiredAttributeError
from services import Service

import idna
import json

import sqlalchemy
from sqlalchemy import Column, func, inspect, select, ForeignKey
from sqlalchemy.dialects.mysql import DATE, INTEGER, TEXT, TINYINT, VARCHAR
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import column_property, validates, relationship, selectinload
from sqlalchemy.types import TypeDecorator


class Orgs(DataModel, DB.Base):
    __tablename__ = "orgs"

    ID = Column("id", INTEGER(10, unsigned=True), unique=True, primary_key=True, nullable=False)
    name = Column("name", VARCHAR(32), nullable=False)
    description = Column("description", VARCHAR(128))

    domains = relationship("Domains", back_populates="org")

    _dictmapping_ = ((Id(), Text("name", flags="patch")), (Text("description", flags="patch"),),
                     (RefProp("domains", flags="patch"),))

    def fromdict(self, patches, *args, **kwargs):
        domains = patches.pop("domains", None)
        DataModel.fromdict(self, patches, *args, **kwargs)
        if domains is not None:
            if self.ID is not None:
                sync = {"synchronize_session": "fetch"}
                Domains.query.filter(Domains.orgID == self.ID, Domains.ID.notin_(domains)).update({Domains.orgID: 0}, **sync)
                Domains.query.filter(Domains.ID.in_(domains)).update({Domains.orgID: self.ID}, **sync)
            else:
                domains = Domains.query.filter(Domains.ID.in_(domains))
                for domain in domains:
                    domain.org = self

    @validates("name")
    def validateName(self, key, value, *args):
        if Orgs.query.filter(Orgs.ID != self.ID, Orgs.name == value).count():
            raise ValueError("Organization '{}' already exists".format(value))
        return value


# Introduced in n109
class OrgParam(DB.Base):
    __tablename__ = "orgparam"

    orgID = Column("org_id", INTEGER(10, unsigned=True), ForeignKey(Orgs.ID), primary_key=True)
    key = Column("key", VARCHAR(32), nullable=False, primary_key=True)
    value = Column("value", VARCHAR(255))

    def __init__(self, orgID, key, value):
        self.orgID = orgID
        self.key = key
        self.value = value

    @classmethod
    def loadLdap(cls, orgID):
        def _addIfDef(dc, d, sc, s, all=False, type=None):
            def tf(v):
                return v if type is None else type(v)
            if s in sc:
                dc[d] = tf(sc[s]) if not all else [tf(v) for v in sc[s].split(",")]

        entries = cls.query.filter(cls.orgID == orgID, cls.key.like("ldap_%")).with_entities(cls.key, cls.value).all()
        if len(entries) == 0:
            return None
        plain = {entry.key: entry.value for entry in entries}

        config = {"connection": {}, "groups": {}, "users": {}}
        _addIfDef(config, "disabled", plain, "ldap_disabled", type=lambda x: x.lower() in ("true", "yes", "1"))
        _addIfDef(config["connection"], "server", plain, "ldap_uri")
        _addIfDef(config["connection"], "bindUser", plain, "ldap_binddn")
        _addIfDef(config["connection"], "bindPass", plain, "ldap_bindpw")
        _addIfDef(config["connection"], "starttls", plain, "ldap_start_tls", type=lambda x: x.lower() in ("true", "yes", "1"))
        _addIfDef(config["connection"], "connections", plain, "data_connections", type=int)
        _addIfDef(config, "baseDn", plain, "ldap_basedn")
        _addIfDef(config, "objectID", plain, "ldap_object_id")
        _addIfDef(config["users"], "username", plain, "ldap_mail_attr")
        _addIfDef(config["users"], "filter", plain, "ldap_user_filter")
        _addIfDef(config["users"], "contactFilter", plain, "ldap_contact_filter")
        _addIfDef(config["users"], "searchAttributes", plain, "ldap_user_search_attrs", all=True)
        _addIfDef(config["users"], "displayName", plain, "ldap_user_displayname")
        _addIfDef(config["users"], "defaultQuota", plain, "ldap_user_default_quota", type=int)
        _addIfDef(config["users"], "templates", plain, "ldap_user_templates", all=True)
        _addIfDef(config["users"], "aliases", plain, "ldap_user_aliases")
        _addIfDef(config["groups"], "groupaddr", plain, "ldap_group_addr")
        _addIfDef(config["groups"], "groupfilter", plain, "ldap_group_filter")
        _addIfDef(config["groups"], "groupname", plain, "ldap_group_name")
        _addIfDef(config["groups"], "groupMemberAttr", plain, "ldap_group_memberof")
        if "ldap_user_attributes" in plain:
            config["users"]["attributes"] = {entry.split(" ", 1)[0]: entry.split(" ", 1)[1]
                                             for entry in plain.getall("ldap_user_attributes") if " " in entry}
        return config

    @classmethod
    def saveLdap(cls, orgID, config):
        def _addIfDef(dc, d, sc, s):
            if s in sc and sc[s] is not None:
                dc[d] = ",".join(str(x) for x in sc[s]) if isinstance(sc[s], list) else str(sc[s])

        flat = {}
        _addIfDef(flat, "ldap_disabled", config, "disabled")
        if "connection" in config:
            _addIfDef(flat, "ldap_uri", config["connection"], "server")
            _addIfDef(flat, "ldap_binddn", config["connection"], "bindUser")
            _addIfDef(flat, "ldap_bindpw", config["connection"], "bindPass")
            _addIfDef(flat, "ldap_start_tls", config["connection"], "starttls")
        _addIfDef(flat, "ldap_basedn", config, "baseDn")
        _addIfDef(flat, "ldap_object_id", config, "objectID")
        if "users" in config:
            _addIfDef(flat, "ldap_mail_attr", config["users"], "username")
            _addIfDef(flat, "ldap_user_displayname", config["users"], "displayName")
            _addIfDef(flat, "ldap_user_filter", config["users"], "filter")
            _addIfDef(flat, "ldap_contact_filter", config["users"], "contactFilter")
            _addIfDef(flat, "ldap_user_search_attrs", config["users"], "searchAttributes")
            _addIfDef(flat, "ldap_user_default_quota", config["users"], "defaultQuota")
            _addIfDef(flat, "ldap_user_templates", config["users"], "templates")
            _addIfDef(flat, "ldap_user_aliases", config["users"], "aliases")
            if "attributes" in config["users"] and config["users"]["attributes"]:
                flat["ldap_user_attributes"] = ["{} {}".format(key, value)
                                                for key, value in config["users"]["attributes"].items()]
        if "groups" in config:
            _addIfDef(flat, "ldap_group_addr", config["groups"], "groupaddr")
            _addIfDef(flat, "ldap_group_filter", config["groups"], "groupfilter")
            _addIfDef(flat, "ldap_group_name", config["groups"], "groupname")
            _addIfDef(flat, "ldap_group_memberof", config["groups"], "groupMemberAttr")
        cls.query.filter(cls.orgID == orgID).delete(synchronize_session=False)
        for key, value in flat.items():
            DB.session.add(OrgParam(orgID, key, value))
        DB.session.commit()

    @classmethod
    def wipeLdap(cls, orgID):
        cls.query.filter(cls.orgID == orgID, cls.key.like("ldap_%")).delete(synchronize_session=False)
        DB.session.commit()

    @classmethod
    def ldapOrgs(cls):
        return [entry[0] for entry in cls.query.with_entities(cls.orgID.distinct()).filter(cls.key.like("ldap_%"))]


class Domains(DataModel, DB.Base, NotifyTable):
    class DomainName(TypeDecorator):
        """Custom column type to allow comparisons to work with unicode names."""
        impl = VARCHAR
        cache_ok = True

        def process_bind_param(self, value, dialect):
            try:
                xfix = "%" if value[0] == "%" else "", "%" if value[-1] == "%" else ""
                return xfix[0]+idna.encode(value.strip("%")).decode("ascii")+xfix[1]
            except Exception:
                try:
                    return value.encode("ascii")
                except Exception:
                    return None

    __tablename__ = "domains"

    ID = Column("id", INTEGER(10, unsigned=True), unique=True, primary_key=True, nullable=False)
    orgID = Column("org_id", INTEGER(10, unsigned=True), ForeignKey(Orgs.ID), nullable=False, server_default="0", index=True)
    _domainname = Column("domainname", DomainName(64), nullable=False)
    homeserverID = OptionalC(105, "0", Column("homeserver", TINYINT(unsigned=True), nullable=False, server_default="0"))
    homedir = Column("homedir", VARCHAR(128), nullable=False, server_default="")
    maxUser = Column("max_user", INTEGER(10, unsigned=True), nullable=False)
    title = Column("title", VARCHAR(128), nullable=False, server_default="")
    address = Column("address", VARCHAR(128), nullable=False, server_default="")
    adminName = Column("admin_name", VARCHAR(32), nullable=False, server_default="")
    tel = Column("tel", VARCHAR(64), nullable=False, server_default="")
    endDay = Column("end_day", DATE, nullable=False, default="3333-03-03")
    domainStatus = Column("domain_status", TINYINT, nullable=False, server_default="0")
    chatID = OptionalC(79, "NULL", Column("chat_id", VARCHAR(26)))
    _syncPolicy = OptionalC(77, "NULL", Column("sync_policy", TEXT))

    org = relationship(Orgs, back_populates="domains")
    homeserver = OptionalNC(105, None,
                            relationship("Servers", foreign_keys=homeserverID, primaryjoin="Domains.homeserverID==Servers.ID"))

    _dictmapping_ = ((Id(), Text("domainname", flags="init"), "displayname"),
                     (Id("orgID", flags="patch"),
                      Int("maxUser", flags="patch"),
                      Int("activeUsers"),
                      Int("inactiveUsers"),
                      Text("title", flags="patch"),
                      Text("address", flags="patch"),
                      Text("adminName", flags="patch"),
                      Text("tel", flags="patch"),
                      Date("endDay", flags="patch"),
                      Int("domainStatus", flags="patch", filter="set")),
                     ({"attr": "syncPolicy", "flags": "patch"},
                      {"attr": "chat", "flags": "patch"},
                      RefProp("homeserver", "homeserverID", flags="patch", filter="set", qopt=selectinload)),
                     (Text("homedir", match=False, flags="hidden"),))

    NORMAL = 0
    SUSPENDED = 1
    OUTOFDATE = 2
    DELETED = 3

    _team = None

    def __init__(self, props: dict, *args, **kwargs):
        if "password" in props:
            self.password = props.pop("password")
        DataModel.__init__(self, props, args, kwargs)

    def fromdict(self, patches, *args, **kwargs):
        DataModel.fromdict(self, patches, args, kwargs)
        if self.chatID:
            with Service("chat", errors=Service.SUPPRESS_INOP) as chat:
                self._team = chat.updateTeam(self)

    @property
    def syncPolicy(self):
        try:
            return json.loads(self._syncPolicy)
        except Exception:
            return None

    @syncPolicy.setter
    def syncPolicy(self, value):
        self._syncPolicy = json.dumps(value, separators=(",", ":")) if value is not None else None

    @hybrid_property
    def domainname(self):
        return self._domainname

    @domainname.setter
    def domainname(self, value):
        try:
            idn = idna.encode(value).decode("ascii")
        except Exception:
            idn = value
        if not formats.domain.match(idn):
            raise ValueError("'{}' is not a valid domain name".format(idn))
        self._domainname = idn

    @property
    def chat(self):
        if not self.chatID:
            return False
        if self._team is None:
            with Service("chat", errors=Service.SUPPRESS_ALL) as grochat:
                self._team = grochat.getTeam(self.chatID)
        return self._team["delete_at"] == 0 if self._team else False

    @chat.setter
    def chat(self, value):
        if value == self.chat:
            return
        if not DB.minVersion(79):
            raise ValueError("Cannot activate chat - please upgrade database schema to at least 79")
        if value and self.domainStatus:
            raise ValueError("Cannot activate chat for deactivated domain")
        with Service("chat") as chat:
            if isinstance(value, str):
                tmp = chat.getTeam(value)
                self.chatID = value
                self._team = tmp
            if self._team:
                tmp = chat.activateTeam(self, value)
                if tmp:
                    self._team["delete_at"] = not value
            else:
                tmp = chat.createTeam(self)
                if tmp:
                    self._team = tmp

    @property
    def displayname(self):
        return idna.decode(self._domainname)

    @validates("_syncPolicy")
    def triggerSyncPolicyUpdate(self, key, value, *args):
        if value != self._syncPolicy:
            users = ["grommunio-sync:policycache-"+user.username
                     for user in Users.query.with_entities(Users.username).filter(Users.domainID == self.ID)]
            if len(users) > 0:
                with Service("redis", errors=Service.SUPPRESS_INOP) as r:
                    r.delete(*users)
        return value

    @staticmethod
    def checkCreateParams(data):
        if "maxUser" not in data:
            return "Missing required property maxUser"

    def delete(self):
        from .users import Users
        self.domainStatus = self.DELETED
        Users.query.filter(Users.domainID == self.ID)\
                   .update({Users.addressStatus: Users.addressStatus.op("&")(0xF) + (self.DELETED << 4)},
                           synchronize_session=False)

    def recover(self):
        from .users import Users
        self.domainStatus = self.NORMAL
        Users.query.filter(Users.domainID == self.ID)\
                   .update({Users.addressStatus: Users.addressStatus.op("&")(0xF) + (self.NORMAL << 4)},
                           synchronize_session=False)

    def purge(self, deleteFiles=False, printStatus=False):
        from .misc import DBConf
        from .mlists import MLists, Associations, Specifieds
        from .roles import AdminRoles as AR, AdminRolePermissionRelation as ARPR
        from .users import Users, Aliases
        users = Users.query.filter(Users.domainID == self.ID)
        if deleteFiles:
            from shutil import rmtree
            us = users.with_entities(Users.maildir).all()
            if printStatus:
                print("Deleting user directories...", end="")
            for user in us:
                if user.maildir != "":
                    rmtree(user.maildir, True)
            if printStatus:
                print("Done.\nDeleting domain directory...", end="")
            rmtree(self.homedir, True)
            if printStatus:
                print("Done.")
        nosync = {"synchronize_session": False}
        mlists = MLists.query.filter(MLists.domainID == self.ID)
        Specifieds.query.filter(Specifieds.listID.in_(mlists.with_entities(MLists.ID))).delete(**nosync)
        Associations.query.filter(Associations.listID.in_(mlists.with_entities(MLists.ID))).delete(**nosync)
        mlists.delete(**nosync)
        Aliases.query.filter(Aliases.mainname.in_(users.with_entities(Users.username))).delete(**nosync)
        users.delete(**nosync)
        permissions = ARPR.query.filter(ARPR.permission == "DomainAdmin", ARPR._params == self.ID)
        roles = []
        for permission in permissions:
            DB.session.delete(permission)
            roles.append(permission.roleID)
        roles = AR.query.filter(AR.ID.in_(roles))
        for role in roles:
            if len(role.permissions) == 0:
                DB.session.delete(role)
        DBConf.query.filter(DBConf.service == "grommunio-admin", DBConf.file == "defaults-domain-"+str(self.ID))\
                    .delete(**nosync)
        DB.session.delete(self)

    @staticmethod
    def create(props, createRole=True, *args, **kwargs):
        def rolename(ID, name):
            maxlen = 32
            base = "Domain Admin ({}/{})"
            sublen = len(str(ID))+len(name)
            if len(base)+sublen-4 > maxlen:
                name = name[:maxlen-sublen-len(base)+3]+"…"
            return base.format(ID, name)

        from .roles import AdminRoles
        from orm.misc import Servers
        from tools.storage import DomainSetup
        from tools.misc import AutoClean
        error = Domains.checkCreateParams(props)
        chat = props.pop("chat", None)
        if error is not None:
            return error, 400
        try:
            domain = Domains(props, *args, **kwargs)
        except (MissingRequiredAttributeError, InvalidAttributeError, MismatchROError, ValueError) as err:
            return err.args[0], 400
        try:
            with AutoClean(lambda: DB.session.rollback()):
                DB.session.add(domain)
                DB.session.flush()
                domain.homeserverID, domain.homedir = Servers.allocDomain(domain.ID, props.get("homeserver"))
                with DomainSetup(domain, DB.session) as ds:
                    ds.run()
                if not ds.success:
                    return "Error during domain setup: "+ds.error, ds.errorCode
                if chat:
                    domain.chat = chat
                DB.session.commit()
            domainAdminRoleName = rolename(domain.ID, domain.domainname)
            if createRole and AdminRoles.query.filter(AdminRoles.name == domainAdminRoleName).count() == 0:
                DB.session.add(AdminRoles({"name": domainAdminRoleName,
                                           "description": "Domain administrator for "+domain.domainname,
                                           "permissions": [{"permission": "DomainAdmin", "params": domain.ID}]}))
                DB.session.commit()
            return domain, 201
        except IntegrityError as err:
            return "Object violates database constraints ({})".format(err.orig.args[1]), 400

    @validates("homeserverID")
    def checkHomeserver(self, key, value, *args):
        from tools.config import Config
        if self.homeserverID and value != self.homeserverID and Config["options"].get("serverExplicitMount"):
            raise ValueError("Cannot change homeserver with explicitly mounted home-directories")
        from .misc import Servers
        if value and Servers.query.filter(Servers.ID == value).count() == 0:
            raise ValueError("Invalid homeserver")
        return value or 0

    @classmethod
    def _commit(cls, *args, **kwargs):
        with Service("systemd", errors=Service.SUPPRESS_ALL) as sysd:
            sysd.reloadService("gromox-delivery.service", "gromox-delivery-queue.service",
                               "gromox-http.service")


from .users import Users
from . import misc

if sqlalchemy.__version__.split(".") >= ["1", "4"]:
    inspect(Domains).add_property("activeUsers",
                                  column_property(select([func.count(Users.ID)])
                                                  .where((Users.domainID == Domains.ID) & (Users.addressStatus == 0))
                                                  .scalar_subquery()))
    inspect(Domains).add_property("inactiveUsers",
                                  column_property(select([func.count(Users.ID)])
                                                  .where((Users.domainID == Domains.ID) & (Users.addressStatus != 0))
                                                  .scalar_subquery()))
else:
    inspect(Domains).add_property("activeUsers",
                                  column_property(select([func.count(Users.ID)])
                                                  .where((Users.domainID == Domains.ID) & (Users.addressStatus == 0))
                                                  .as_scalar()))
    inspect(Domains).add_property("inactiveUsers",
                                  column_property(select([func.count(Users.ID)])
                                                  .where((Users.domainID == Domains.ID) & (Users.addressStatus != 0))
                                                  .as_scalar()))

Domains.NTregister()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import logging


class ServiceUnavailableError(Exception):
    """Service not available.

    Thrown to indicate that the service is not available,
    either because initialization failed or an error occurred
    that is typically caused by the external service being unreachable
    (in contrast to errors caused by e.g. faulty data).

    When thrown in a service constructor to indicate that reloading
    at a later time may resolve the issue
    (putting the service in SUSPENDED state)."""
    pass


class ServiceDisabledError(Exception):
    """Service is manually disabled.

    Should be thrown in service constructor to indicate
    that the service is theoretically available,
    but was disabled manually and can be reactivated later."""
    pass


class InstanceDefault(Exception):
    """Service parametrization not supported.

    Thrown in the service constructor to indicate that parametrization is
    either not supported or the given parameters would offer the same
    functionality the default instance does."""


class _ServiceHubMeta(type):
    def __contains__(cls, value):
        return value in cls._services

    def __iter__(cls):
        return iter(cls._services.values())

    def __getitem__(cls, item):
        return cls._services[item]


class ServiceHub(metaclass=_ServiceHubMeta):
    _services = {}
    _instances = {}

    UNINITIALIZED = -1  # State has not been set
    LOADED = 0          # Service is available
    UNAVAILABLE = 1     # Temporarily unavailable, might become available automatically
    SUSPENDED = 2       # Needs to be reloaded in order to work
    ERROR = 3           # Fatal error
    DISABLED = 4        # Ok, but disabled manually

    _names = {UNINITIALIZED: "UNLOADED",
              LOADED: "LOADED",
              UNAVAILABLE: "UNAVAILABLE",
              SUSPENDED: "SUSPENDED",
              ERROR: "ERROR",
              DISABLED: "DISABLED"}

    class ServiceInfo:
        def __init__(self, name, mgrclass, exchandler, maxreloads, maxfailures, reloadlocktime, argspec, argname):
            self._argname = argname
            self._argspec = argspec
            self._maxfailures = maxfailures
            self._maxreloads = maxreloads
            self._name = name
            self._reloadlocktime = reloadlocktime
            self.exchandler = exchandler
            self.logger = logging.getLogger(self._name)
            self.mgrclass = mgrclass

        def checkArgs(self, *args):
            for spec in self._argspec:
                if len(spec) != len(args):
                    continue
                try:
                    return tuple(t(a) for t, a in zip(spec, args))
                except Exception:
                    pass
            raise ValueError("Invalid service parameters")

    class ServiceInstance:
        def __init__(self, service, *args):
            self._args = args
            self._failures = 0
            self._lastreload = 0
            self._name = ServiceHub.servicename(service._name, *args)
            self._reloads = 0
            self._service = service
            self._state = ServiceHub.UNINITIALIZED
            self.exc = None
            self.logger = service.logger.getChild("".join("{:02x}".format(x) for x in self._name.encode("utf8")))
            self.logger.name = self.name
            self.manager = None
            self.load()

        def __repr__(self):
            statename = {-1: "UNINITIALIZED", 0: "LOADED", 1: "UNAVAILABLE", 2: "SUSPENDED", 3: "ERROR"}
            return "<Service '{}' state {}>".format(self.name, statename.get(self._state))

        def _checkArgs(self):
            return self._service.checkArgs(*self._args)

        def disable(self):
            self.state = ServiceHub.DISABLED
            self.exc = ServiceDisabledError("Service disabled manually")

        def failed(self, newstate, exception):
            self._failures += 1
            self.exc = exception
            if self._service._maxfailures is not None and self._failures > self._service._maxfailures:
                self.state = ServiceHub.ERROR
                self._failures = self._service._maxfailures
                self.logger.info("Service deactivated after too many errors")
                return False
            self.state = newstate
            return True

        def load(self, force_reload=False):
            from time import time
            if (self._state not in (ServiceHub.UNINITIALIZED, ServiceHub.SUSPENDED) or
               time()-self._lastreload < self._service._reloadlocktime) and not force_reload:
                return
            self._reloads += 1
            try:
                self._checkArgs()
                self.manager = self._service.mgrclass(*self._args)
                self.state = ServiceHub.LOADED
                self._reloads = 0
                self.logger.info("Service loaded successfully")
                return
            except ServiceUnavailableError as err:
                self.exc = err
                self.logger.warning("Failed to load service: "+" - ".join(str(arg) for arg in err.args))
                self.state = ServiceHub.SUSPENDED if self._reloads <= self._service._maxreloads else ServiceHub.ERROR
                self._reloads = min(self._reloads, self._service._maxreloads)
            except ServiceDisabledError as err:
                self.exc = err
                self.logger.warning("Failed to load service: "+err.args[0])
                self.state = ServiceHub.DISABLED
            except InstanceDefault:
                raise
            except Exception as err:
                self.exc = err
                self.logger.error("Failed to load service: "+" - ".join(str(arg) for arg in err.args))
                self.state = ServiceHub.ERROR
            self._lastreload = time()
            self.manager = None

        @property
        def available(self):
            return self.state not in (ServiceHub.SUSPENDED, ServiceHub.ERROR, ServiceHub.DISABLED)

        @property
        def failures(self):
            return self._failures

        @property
        def maxfailures(self):
            return self._service._maxfailures

        @property
        def maxreloads(self):
            return self._service._maxreloads

        @property
        def name(self):
            return self._name

        @property
        def reloads(self):
            return self._reloads

        @property
        def state(self):
            return self._state

        @state.setter
        def state(self, value):
            newstate = value if ServiceHub.UNINITIALIZED <= value <= ServiceHub.DISABLED else ServiceHub.ERROR
            if newstate != self._state:
                self.logger.debug("State changed {} -> {}".format(self.statename, ServiceHub.statename(newstate)))
            self._state = newstate

        @property
        def statename(self):
            return ServiceHub.statename(self._state)

    @classmethod
    def register(cls, name, exchandler=lambda *args, **kwargs: None, maxreloads=0, maxfailures=None, reloadlocktime=1,
                 argspec=((),), argname=None):
        """Decorator to register a new service provider class.

        Register a new class as service provider.

        An exception handler function can be defined taking the service instance and exception raised in the service context,
        returning either None (indicating that the exception was not handled) or the new service state and optionally
        an error message (indicating that the original exception was handled, throwing a ServiceUnavailableException instead).
        This provides a convenient way to react to exception resulting from lost connections and put automatic reloading and
        deactivation mechanics to use.

        `maxreloads` and `maxfailures` provide limits which, when reached, will permanently disable a service
        (the service can still be recovered with force_reload).
        `maxreloads` is the maximum number of automatic reloads, i.e. number of failed load operations.
        `maxfailures` is the maximum number of exceptions handled by the exception handler

        `reloadlocktime` can be used for rate-limiting automatic reloads. This can prevent that a short temporary downtime
        of a service permanently disables it because it is requested multiple times in rapid succession.

        Parameters
        ----------
        name : str
            Name of the service.
        exchandler : function, optional
            Exception handler. The default is the empty function.
        **kwargs : None
            Keyword argument dict, only for compatibility
        maxreloads : int, optional
            Number of acceptable automatic reloads before service enters ERROR state. The default is 0.
        maxfailures : int, optional
            Number acceptable load error before service enters ERROR state or None for indefinite reloading.
            The default is None.
        reloadlocktime : float, optional
            Minimum time (in seconds) between automatic reloads. The default is 1.
        argspec : tuple of tuples of types, optional
            Allowed types for arguments. Default is `((),)` (no arguments).
        argname : function, optional
            Function returning a sensible name for given arguments, or None for fallback naming. Default is None.
        """
        def inner(mgrclass):
            cls._services[name] = cls.ServiceInfo(name, mgrclass, exchandler, maxreloads, maxfailures, reloadlocktime,
                                                  argspec, argname)
            return mgrclass
        return inner

    @classmethod
    def load(cls, service, *args, force_reload=False):
        if service not in cls._services:
            raise ServiceUnavailableError("Service '{}' does not exist.".format(service))

        args = cls._services[service].checkArgs(*args)
        instanceKey = (service, *args)
        if force_reload or instanceKey not in cls._instances:
            try:
                cls._instances[instanceKey] = cls.ServiceInstance(cls._services[service], *args)
            except InstanceDefault:
                cls._instances[instanceKey] = cls.load(service)
        else:
            cls._instances[instanceKey].load()
        return cls._instances[instanceKey]

    @classmethod
    def unload(cls, instance):
        servicekey = (instance._service._name, *instance._args)
        if servicekey in cls._services:
            del cls._services[servicekey]

    @classmethod
    def statename(cls, state):
        return cls._names.get(state, "UNKNOWN")

    @classmethod
    def services(cls):
        return [service for service in cls._services]

    @classmethod
    def instances(cls, service=None):
        return [(key[1:], value) for key, value in cls._instances.items() if service is None or key[0] == service]

    @classmethod
    def servicename(cls, service, *args):
        base = service+"@"
        if service in cls._services and cls._services[service]._argname:
            cls._services[service].checkArgs(*args)
            name = cls._services[service]._argname(*args)
            if name is not None:
                return base+name
        base += "default" if not args else repr(args[0]) if len(args) == 1 else "["+",".join(repr(arg) for arg in args)+"]"
        return base


class Service:
    SUPPRESS_NONE = 0   # All exceptions are passed to the calling function
    SUPPRESS_INOP = 1   # Suppress exceptions indicating service unavailability
    SUPPRESS_ALL = 2    # Suppress all exceptions

    class Stub:
        def __init__(self, name):
            self.name = name

        def __getattr__(self, name):
            raise ServiceUnavailableError("Service '{}' is currently not available".format(self.name))

    def __init__(self, name, *args, errors=SUPPRESS_NONE):
        displayname = ServiceHub.servicename(name, *args)
        try:
            self.__suppress = errors
            self.__service = ServiceHub.load(name, *args)
            self.plugin = self.__service._service.mgrclass
            self.__mgr = self.__service.manager if self.__service.available else self.Stub(displayname)
        except Exception:
            if errors == self.SUPPRESS_ALL:
                self.__service = None
                self.__mgr = self.Stub(displayname)
            else:
                raise

    def __enter__(self):
        return self.__mgr

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None or isinstance(exc_value, ServiceUnavailableError):
            return self.__suppress != 0
        excresult = self.__service._service.exchandler(self.__service, exc_value)
        if isinstance(excresult, tuple):
            newstate, msg = excresult
        else:
            newstate, msg = excresult, "Service '{}' is currently not available".format(self.__service.name)
        if newstate == 0:
            return True
        if newstate:
            self.__service.failed(newstate, exc_value)
            if self.__suppress:
                return True
            raise ServiceUnavailableError(msg)
        if self.__suppress == Service.SUPPRESS_ALL:
            return True

    @staticmethod
    def available(name, *args):
        return name in ServiceHub and ServiceHub.load(name, *args).available

    def service(self):
        return self.__mgr


from . import chat, exmdb, ldap, redis, systemd



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 grommunio GmbH

import json
import subprocess

from flask import jsonify, request

import api

from api.core import API, secure
from api.security import checkPermissions
from tools.permissions import SystemAdminROPermission, SystemAdminPermission


@API.route(api.BaseRoute+"/system/mailq", methods=["GET"])
@secure()
def getMailqData():
    checkPermissions(SystemAdminROPermission())
    try:
        postfixMailq = subprocess.run("mailq", stdout=subprocess.PIPE, universal_newlines=True).stdout
    except Exception as err:
        API.logger.error("Failed to run mailq: {} ({})"
                         .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)))
        postfixMailq = "Failed to run mailq."
    try:
        gromoxMailq = subprocess.run("gromox-mailq", stdout=subprocess.PIPE, universal_newlines=True).stdout
    except Exception as err:
        API.logger.error("Failed to run gromox-mailq: {} ({})"
                         .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)))
        gromoxMailq = ""
    try:
        postqueue = subprocess.run(["postqueue", "-j"], stdout=subprocess.PIPE, universal_newlines=True).stdout
        postqueue = [json.loads(line) for line in postqueue.split("\n") if line]
    except Exception as err:
        API.logger.error("Failed to run postqueue: {} ({})"
                         .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)))
        postqueue = []
    return jsonify(postfixMailq=postfixMailq, gromoxMailq=gromoxMailq, postqueue=postqueue)


@API.route(api.BaseRoute+"/system/mailq/flush", methods=["POST"])
@secure()
def flushMailq():
    checkPermissions(SystemAdminPermission())
    if "queue" not in request.args:
        return jsonify(message="Missing queue parameter"), 400
    from subprocess import PIPE
    target = request.args.get("queue")
    result = subprocess.run(["postqueue", "-i", target], stdout=PIPE, stderr=PIPE, universal_newlines=True)
    log = API.logger.warning if result.returncode else API.logger.info
    if result.stdout:
        log("Postqueue (out): "+result.stdout)
    if result.stderr:
        log("Postqueue (err): "+result.stderr)
    if result.returncode:
        return jsonify(message="Call to postqueue failed ({})".format(result.returncode)), 500
    return jsonify(message="Success")


def postsuper(op):
    from subprocess import PIPE
    targets = request.args.get("queue", "ALL").split(",")
    command = ["sudo", "postsuper"]+[t for param in ((op, target) for target in targets) for t in param]
    result = subprocess.run(command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    log = API.logger.warning if result.returncode else API.logger.info
    if result.stdout:
        log("Postsuper (out): "+result.stdout)
    if result.stderr:
        log("Postsuper (err): "+result.stderr)
    return result


@API.route(api.BaseRoute+"/system/mailq/delete", methods=["POST"])
@secure()
def deleteMailq():
    checkPermissions(SystemAdminPermission())
    result = postsuper("-d")
    if result.returncode:
        return jsonify(message="Call to postsuper failed ({})".format(result.returncode)), 500
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/system/mailq/requeue", methods=["POST"])
@secure()
def requeueMailq():
    checkPermissions(SystemAdminPermission())
    result = postsuper("-r")
    if result.returncode:
        return jsonify(message="Call to postsuper failed ({})".format(result.returncode)), 500
    return jsonify(message="Success")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 grommunio GmbH

import shutil

from flask import jsonify, request

import api
from api.core import API, secure
from api.security import checkPermissions

from services import Service, ServiceUnavailableError
from tools.ldap import downsyncObject, importObject
from tools.permissions import SystemAdminPermission, SystemAdminROPermission, DomainAdminPermission, DomainAdminROPermission
from tools.permissions import OrgAdminPermission
from tools.tasq import TasQServer

from orm import DB


def _getTarget():
    """Get LDAP target organization and domains

    Returns
    -------
    int
        Organization ID or None if not found
    list
        List of domain objects containing ID and domainname
    """
    from orm.domains import Domains
    if "domain" in request.args:
        domainID = int(request.args["domain"])
        checkPermissions(DomainAdminROPermission(domainID))
        domain = Domains.query.filter(Domains.ID == domainID)\
                        .with_entities(Domains.ID, Domains.orgID, Domains.domainname).first()
        return (domain.orgID, [domain]) if domain else (None, [])
    elif "organization" in request.args:
        orgID = int(request.args["organization"])
        checkPermissions(OrgAdminPermission(orgID))
        domains = Domains.query.filter(Domains.orgID == orgID).with_entities(Domains.ID, Domains.domainname).all()
        return orgID, domains
    return None, []


@API.route(api.BaseRoute+"/domains/ldap/search", methods=["GET"])
@secure(requireDB=True, authLevel="user")
def searchLdap():
    if "domain" in request.args:
        checkPermissions(DomainAdminROPermission(int(request.args["domain"])))
    elif "organization" in request.args:
        checkPermissions(OrgAdminPermission(int(request.args["organization"])))
    orgID, domains = _getTarget()
    if len(domains) == 0:
        return jsonify(message="Invalid domain or organization"), 400
    limit = int(request.args.get("limit", 50))
    domainnames = [domain.domainname for domain in domains]
    with Service("ldap", orgID) as ldap:
        ldapusers = ldap.searchUsers(request.args.get("query"), domainnames, limit=limit or None,
                                     filterIncomplete=request.args.get("showAll") != "true")
    return jsonify(data=[{"ID": ldap.escape_filter_chars(u.ID), "name": u.name, "email": u.email,
                          "type": u.type, "error": u.error} for u in ldapusers if u.ID])


def ldapDownsync(orgID=None, domainID=None):
    params = {"domainID": domainID} if domainID else {"orgID": orgID} if orgID is not None else {}
    params["lang"] = request.args.get("lang", "")
    params["import"] = request.args.get("import") == "true"
    permission = DomainAdminROPermission(domainID) if domainID else \
        OrgAdminPermission(orgID) if orgID else SystemAdminROPermission()
    task = TasQServer.create("ldapSync", params, permission)
    timeout = float(request.args.get("timeout", 1))
    if timeout > 0:
        TasQServer.wait(task.ID, timeout)
    if not task.done:
        return jsonify(message="Created background task #"+str(task.ID), taskID=task.ID), 202
    if task.state == task.COMPLETED:
        return jsonify(message=task.message, data=task.params.get("result", []))
    return jsonify(message="Synchronization failed: "+task.message), 500


@API.route(api.BaseRoute+"/domains/ldap/downsync", methods=["POST"])
@secure(requireDB=True, authLevel="user")
def ldapDownsyncAll():
    checkPermissions(SystemAdminPermission())
    return ldapDownsync()


@API.route(api.BaseRoute+"/domains/<int:domainID>/ldap/downsync", methods=["POST"])
@secure(requireDB=True, authLevel="user")
def ldapDownsyncDomain(domainID):
    checkPermissions(DomainAdminPermission(domainID))
    return ldapDownsync(domainID=domainID)


@API.route(api.BaseRoute+"/domains/ldap/importUser", methods=["POST"])
@secure(requireDB=True, authLevel="user")
def importLdapUser():
    checkPermissions(DomainAdminPermission("*"))
    from services.ldap import LdapService
    orgID, domains = _getTarget()
    ldap = Service("ldap", orgID).service()

    if "ID" not in request.args:
        return jsonify(message="Missing ID"), 400
    try:
        ID = LdapService.unescapeFilterChars(request.args["ID"])
    except Exception:
        return jsonify(message="Invalid ID"), 400

    force = request.args.get("force")
    lang = request.args.get("lang", "")
    userinfo = ldap.getUserInfo(ID)
    if userinfo is None:
        return jsonify(message="LDAP object not found"), 404

    result, code = importObject(userinfo, ldap, orgID=orgID, force=force, lang=lang, syncMembers=True, syncExisting=True)

    return jsonify(result), code


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/downsync", methods=["PUT"])
@secure(requireDB=True, authLevel="user")
def updateLdapUser(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    message, code = downsyncObject(user)
    return jsonify(user.fulldesc())


@API.route(api.BaseRoute+"/domains/ldap/check", methods=["GET", "DELETE"])
@secure(requireDB=True, authLevel="user")
def checkLdapUsers():
    from orm.users import Users
    orgID, domains = _getTarget()
    readonly = request.method == "GET"
    if orgID is not None:
        Permission = DomainAdminROPermission if readonly else DomainAdminPermission
        for domain in domains:
            checkPermissions(Permission(domain.ID))
        domainFilter = (Users.domainID.in_(domain.ID for domain in domains),)
    else:
        checkPermissions(SystemAdminROPermission() if readonly else SystemAdminPermission())
        domainFilter = ()
    users = Users.query.filter(Users.externID != None, *domainFilter)\
                       .with_entities(Users.ID, Users.username, Users.externID, Users.maildir, Users.orgID)\
                       .all()
    if len(users) == 0:
        return jsonify(message="No LDAP users found", **{"orphaned" if request.method == "GET" else "deleted": []})
    orphaned = []
    for user in users:
        with Service("ldap", user.orgID) as ldap:
            if ldap.getUserInfo(user.externID) is None:
                orphaned.append(user)
    if len(orphaned) == 0:
        return jsonify(message="All LDAP users are valid", **{"orphaned" if request.method == "GET" else "deleted": []})
    orphanedData = [{"ID": user.ID, "username": user.username} for user in orphaned]
    if request.method == "GET":
        return jsonify(orphaned=orphanedData)
    deleteMaildirs = request.args.get("deleteFiles") == "true"
    homeserver = None
    users = Users.query.filter(Users.ID.in_(orphan.ID for orphan in orphaned)).order_by(Users.homeserverID).all()
    index = 0
    while index < len(users):
        try:
            with Service("exmdb") as exmdb:
                if homeserver != users[index].homeserverID:  # Reuse the exmdb client for users on the same server
                    user = users[index]
                    client = exmdb.ExmdbQueries(exmdb.host if user.homeserverID == 0 else user.homeserver.hostname,
                                                exmdb.port, user.maildir, True)
                    homeserver = user.homeserverID
                while index < len(users) and users[index].homeserverID == homeserver:
                    client.unloadStore(users[index].maildir)
                    if deleteMaildirs:
                        shutil.rmtree(users[index].maildir, ignore_errors=True)
                    users[index].delete()
                    index += 1
        except ServiceUnavailableError:
            API.logger.warning("Failed to unload store: exmdb service not available")
            index += 1
    DB.session.commit()
    return jsonify(deleted=orphanedData)


@API.route(api.BaseRoute+"/domains/ldap/dump", methods=["GET"])
@secure(requireDB=True, authLevel="user")
def dumpLdapUsers():
    orgID, domains = _getTarget()
    permissions = request.auth["user"].permissions()
    if not any(DomainAdminROPermission(domain.ID) in permissions for domain in domains):
        return jsonify(message="Insufficient permissions"), 403
    try:
        with Service("ldap", orgID) as ldap:
            ID = ldap.unescapeFilterChars(request.args["ID"])
    except Exception:
        return jsonify(message="Invalid ID"), 400
    ldapuser = ldap.dumpUser(ID)
    if ldapuser is None:
        return jsonify(message="User not found"), 404
    return jsonify(data=str(ldapuser))


@API.route(api.BaseRoute+"/system/orgs/<int:ID>/ldap/downsync", methods=["POST"])
@secure(requireDB=True, authLevel="user")
def ldapDownsyncOrganization(ID):
    checkPermissions(OrgAdminPermission(ID))
    return ldapDownsync(orgID=ID)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from . import DB, logger

from tools.DataModel import DataModel, Id, Date, Int, Text
from tools.misc import RecursiveDict

from sqlalchemy import Column, func, select
from sqlalchemy.dialects.mysql import INTEGER, TINYINT, VARCHAR, TEXT, TIMESTAMP
from sqlalchemy.ext.hybrid import hybrid_property

import json


def _trydec(value, default=None):
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return default


class DBConf(DB.Base):
    __tablename__ = "configs"

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True)
    service = Column("service", VARCHAR(200), nullable=False, index=True)
    file = Column("file", VARCHAR(200), nullable=False, index=True)
    key = Column("key", VARCHAR(200), nullable=False)
    value = Column("value", VARCHAR(200), nullable=False, default="")

    @staticmethod
    def getFile(service, file, structured=False):
        """Read config file to dict.

        Values are json decoded if possible, otherwise retained as strings.

        Parameters
        ----------
        service : str
            Name of the service.
        file : str
            Name of the config file
        structured : bool, optional
            Unpack keys into RecursiveDict. The default is False.

        Returns
        -------
        dict
            Key/value pairs of the file
        """
        entries = DBConf.query.filter(DBConf.service == service, DBConf.file == file).with_entities(DBConf.key, DBConf.value)
        data = {entry.key: _trydec(entry.value, entry.value) for entry in entries}
        return RecursiveDict(data) if structured else data

    @staticmethod
    def getValue(service, file, key, default=None):
        """Get single config value.

        Parameters
        ----------
        service : str
            Service name
        file : str
            File name
        key : str
            Configuration key
        default : any, optional
            Default value to return if the parameter does not exist. The default is None.

        Returns
        -------
        str
            Configuration value.
        """
        entry = DBConf.query.filter(DBConf.service == service, DBConf.file == file, DBConf.key == key)\
                            .with_entities(DBConf.value).first()
        return default if entry is None else entry.value

    @staticmethod
    def setFile(service, file, data):
        """Write key-value mapping to config file.

        Existing contents are removed.

        Values are json encoded before writing.

        If data is an instance of RecursiveDict, it is flattened before writing.

        Parameters
        ----------
        service : str
            Name of the service
        file : str
            Name of the config file
        data : dict
            Key/value pairs to save.
        """
        data = data.flat() if isinstance(data, RecursiveDict) else data
        data = {key: json.dumps(value, separators=(",", ":")) for key, value in data.items()}
        DBConf.query.filter(DBConf.service == service, DBConf.file == file).delete()
        DB.session.bulk_insert_mappings(DBConf, [dict(service=service, file=file, key=key, value=value)
                                                 for key, value in data.items()])


class TasQ(DataModel, DB.Base):
    __tablename__ = "tasq"

    ID = Column("id", INTEGER(10, unsigned=True), primary_key=True)
    command = Column("command", VARCHAR(16))
    state = Column("state", TINYINT(unsigned=True), server_default="0")
    created = Column("created", TIMESTAMP, server_default="now()")
    updated = Column("updated", TIMESTAMP, server_default="now()")
    message = Column("message", VARCHAR(160), server_default="")
    _params = Column("params", TEXT, nullable=True, default="{}")
    access = Column("access", TEXT, nullable=True)

    _dictmapping_ = ((Id(), Text("command", flags="init")),
                     (Id("state"),
                      Date("created", time=True),
                      Date("updated", time=True),
                      Text("message")),
                     ({"attr": "params", "flags": "patch"},))

    @property
    def params(self):
        return json.loads(self._params) if self._params is not None else {}

    @params.setter
    def params(self, value):
        self._params = json.dumps(value, separators=(',', ':'))

    @property
    def permission(self):
        from tools.permissions import Permissions
        return Permissions.load(self.access)

    @permission.setter
    def permission(self, perm):
        from tools.permissions import Permissions
        self.access = Permissions.dump(perm)


class Servers(DataModel, DB.Base):
    __tablename__ = "servers"

    ID = Column("id", TINYINT(unsigned=True), primary_key=True)
    hostname = Column("hostname", VARCHAR(255), nullable=False, unique=True)
    extname = Column("extname", VARCHAR(255), nullable=False, unique=True)

    _dictmapping_ = ((Id(), Text("hostname", flags="patch")),
                     (Text("extname", flags="patch"),),
                     (Int("users"),
                      Int("domains")))

    # Defined as hybrid_property instead of column_property to break cyclical import dependency
    # Domains -> Users -> Servers -> Domains
    @hybrid_property
    def users(self):
        from .users import Users
        return Users.query.filter(Users.homeserverID == self.ID).count()

    @users.expression
    def users(cls):
        from .users import Users
        return select([func.count(Users.ID)]).where(Users.homeserverID == cls.ID).as_scalar()

    @hybrid_property
    def domains(self):
        from .domains import Domains
        return Domains.query.filter(Domains.homeserverID == self.ID).count()

    @domains.expression
    def domains(cls):
        from .domains import Domains
        return select([func.count(Domains.ID)]).where(Domains.homeserverID == cls.ID).as_scalar()

    @staticmethod
    def _getServer(objID, serverID=None, domain=False):
        """Select a server for an object

        Parameters
        ----------
        objID : int
            ID of the object
        serverID : int, optional
            ID of the server or None to select automatically. The default is None.

        Raises
        ------
        ValueError
            Server with specified ID could not be found.

        Returns
        -------
        Servers
            Selected server object or None if no servers are configured (i.e. single server setup)
        """
        if not DB.minVersion(105):
            return None
        from tools.config import Config
        if serverID:
            server = Servers.query.filter(Servers.ID == serverID).first()
            if server is None:
                raise ValueError("Requested server #{} not found".format(serverID))
            return server
        servers = Servers.query.count()
        if servers == 0:
            return None
        policy = DBConf.getValue("grommunio-admin", "multi-server", "policy", default="round-robin")
        if policy == "balanced":
            return Servers.query.order_by((Servers.domains if domain else Servers.users).asc()).first()
        elif policy == "first":
            index = 0
        elif policy == "last":
            index = servers-1
        elif policy == "random":
            import random
            index = random.randint(0, servers-1)
        else:  # default (policy == "round-robin")
            if policy != "round-robin":
                logger.warning("Unknown multi-server policy '{}'. Defaulting to round-robin.".format(policy))
            index = objID % servers
        return Servers.query.order_by(Servers.ID).offset(index).first()

    @staticmethod
    def allocUser(userID, serverID=None):
        """Select a server to store new user on.

        Parameters
        ----------
        userID : int
            ID of the new user
        serverID : int, optional
            Server to use or None for automatic selection. The default is None.

        Returns
        -------
        tuple(int, str)
            2-tuple containing the server ID and path
        """
        from tools.config import Config
        from os import path
        server = Servers._getServer(userID, serverID)
        serverMount = server is not None and Config["options"].get("serverExplicitMount")
        targetPath = Config["options"]["userPrefix"]
        targetPath = path.join(targetPath, server.hostname) if serverMount else targetPath
        return (0 if server is None else server.ID, targetPath)

    @staticmethod
    def allocDomain(domainID, serverID=None):
        """Select a server to store new domain on.

        Parameters
        ----------
        domainID : int
            ID of the new domain
        serverID : int, optional
            Server to use or None for automatic selection. The default is None.

        Returns
        -------
        tuple(int, str)
            2-tuple containing the server ID and path
        """
        from tools.config import Config
        from os import path
        server = Servers._getServer(domainID, serverID, domain=True)
        serverMount = server is not None and Config["options"].get("serverExplicitMount")
        targetPath = Config["options"]["domainPrefix"]
        targetPath = path.join(targetPath, server.hostname) if serverMount else targetPath
        return (0 if server is None else server.ID, targetPath)
